{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Import libraries"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "import ee\n",
    "import geemap"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Create an interactive map"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "Map = geemap.Map(center=[40, -100], zoom=4)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Add Earth Engine Python script"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "# Add Earth Engine dataset\n",
    "image = ee.Image(\"USGS/SRTMGL1_003\")\n",
    "\n",
    "#  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n",
    "#  Chapter:      F5.2 Zonal Statistics\n",
    "#  Checkpoint:   F52b\n",
    "#  Authors:      Sara Winsemius and Justin Braaten\n",
    "#  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n",
    "\n",
    "# Functions that the rest of the chapter is based on.\n",
    "\n",
    "# Returns a function for adding a buffer to points and optionally transforming\n",
    "# to rectangular bounds\n",
    "def bufferPoints(radius, bounds):\n",
    "    return function(pt) {\n",
    "        pt = ee.Feature(pt)\n",
    "        'return bounds ? pt.buffer(radius).bounds()' : pt.buffer(\n",
    "            radius)\n",
    "    }\n",
    "\n",
    "\n",
    "# Reduces images in an ImageCollection by regions defined in a\n",
    "# FeatureCollection. Similar to mapping reduceRegions over an ImageCollection,\n",
    "# but breaks the task up a bit more and includes parameters for managing\n",
    "# property names.\n",
    "def zonalStats(ic, fc, params):\n",
    "    # Initialize internal params dictionary.\n",
    "    _params = {\n",
    "        'reducer': ee.Reducer.mean(),\n",
    "        'scale': None,\n",
    "        'crs': None,\n",
    "        'bands': None,\n",
    "        'bandsRename': None,\n",
    "        'imgProps': None,\n",
    "        'imgPropsRename': None,\n",
    "        'datetimeName': 'datetime',\n",
    "        'datetimeFormat': 'YYYY-MM-dd HH:'mm':ss'\n",
    "    }\n",
    "\n",
    "    # Replace initialized params with provided params.\n",
    "    if (params) {\n",
    "        for param in params:\n",
    "            _params[param] = params[param] || _params[param]\n",
    "\n",
    "    }\n",
    "\n",
    "    # Set default parameters based on an image representative.\n",
    "    imgRep = ic.first()\n",
    "    nonSystemImgProps = ee.Feature(None) \\\n",
    "        .copyProperties(imgRep).propertyNames()\n",
    "    if (!_params.bands) _params.bands = imgRep.bandNames()\n",
    "    if (!_params.bandsRename) _params.bandsRename = _params.bands\n",
    "    if (!_params.imgProps) _params.imgProps = nonSystemImgProps\n",
    "    if (!_params.imgPropsRename) _params.imgPropsRename = _params \\\n",
    "        .imgProps\n",
    "\n",
    "    # Map the reduceRegions function over the image collection.\n",
    "\n",
    "def func_fpj(img):\n",
    "        # Select bands (optionally rename), set a datetime & timestamp property.\n",
    "        img = ee.Image(img.select(_params.bands, _params \\\n",
    "                .bandsRename)) \\\n",
    "            .set(_params.datetimeName, img.date().format(\n",
    "                _params.datetimeFormat)) \\\n",
    "            .set('timestamp', img.get('system:time_start'))\n",
    "\n",
    "        # Define final image property dictionary to set in output features.\n",
    "        propsFrom = ee.List(_params.imgProps) \\\n",
    "            .cat(ee.List([_params.datetimeName,\n",
    "            'timestamp']))\n",
    "        propsTo = ee.List(_params.imgPropsRename) \\\n",
    "            .cat(ee.List([_params.datetimeName,\n",
    "            'timestamp']))\n",
    "        imgProps = img.toDictionary(propsFrom).rename(\n",
    "            propsFrom, propsTo)\n",
    "\n",
    "        # Subset points that intersect the given image.\n",
    "        fcSub = fc.filterBounds(img.geometry())\n",
    "\n",
    "        # Reduce the image by regions.\n",
    "        return img.reduceRegions({\n",
    "                'collection': fcSub,\n",
    "                'reducer': _params.reducer,\n",
    "                'scale': _params.scale,\n",
    "                'crs': _params.crs\n",
    "            }) \\\n",
    "            .map(function(f) {\n",
    "                return f.set(imgProps)\n",
    "            })\n",
    "\n",
    "        # Converts the feature collection of feature collections to a single\n",
    "        #feature collection.\n",
    "\n",
    "    results = ic.map(func_fpj\n",
    ").flatten()\n",
    "\n",
    "\n",
    "\n",
    "\n",
    "\n",
    "\n",
    "\n",
    "\n",
    "\n",
    "\n",
    "\n",
    "\n",
    "\n",
    "\n",
    "\n",
    "\n",
    "\n",
    "\n",
    "\n",
    "\n",
    "\n",
    "\n",
    "\n",
    "\n",
    "\n",
    "\n",
    "\n",
    "\n",
    "\n",
    "\n",
    "\n",
    "\n",
    "\n",
    "\n",
    "\n",
    ").flatten()\n",
    "\n",
    "    return results\n",
    "\n",
    "\n",
    "# Creating points that will be used for the rest of the chapter.\n",
    "# Alternatively, you could load your own points.\n",
    "pts = ee.FeatureCollection([\n",
    "    ee.Feature(ee.Geometry.Point([-118.6010, 37.0777]), {\n",
    "        'plot_id': 1\n",
    "    }),\n",
    "    ee.Feature(ee.Geometry.Point([-118.5896, 37.0778]), {\n",
    "        'plot_id': 2\n",
    "    }),\n",
    "    ee.Feature(ee.Geometry.Point([-118.5842, 37.0805]), {\n",
    "        'plot_id': 3\n",
    "    }),\n",
    "    ee.Feature(ee.Geometry.Point([-118.5994, 37.0936]), {\n",
    "        'plot_id': 4\n",
    "    }),\n",
    "    ee.Feature(ee.Geometry.Point([-118.5861, 37.0567]), {\n",
    "        'plot_id': 5\n",
    "    })\n",
    "])\n",
    "\n",
    "print('Points of interest', pts)\n",
    "\n",
    "#  -----------------------------------------------------------------------\n",
    "#  CHECKPOINT\n",
    "#  -----------------------------------------------------------------------\n",
    "\n",
    "# Example 1: Topographic variables\n",
    "\n",
    "# Buffer the points.\n",
    "ptsTopo = pts.map(bufferPoints(45, False))\n",
    "\n",
    "# Import the MERIT global elevation dataset.\n",
    "elev = ee.Image('MERIT/DEM/v1_0_3')\n",
    "\n",
    "# Calculate slope from the DEM.\n",
    "slope = ee.Terrain.slope(elev)\n",
    "\n",
    "# Concatenate elevation and slope as two bands of an image.\n",
    "topo = ee.Image.cat(elev, slope)\n",
    "    # Computed images do not have a 'system:time_start' property; add one based \\\n",
    "    .set('system:time_start', ee.Date('2000-01-01').millis())\n",
    "\n",
    "# Wrap the single image in an ImageCollection for use in the\n",
    "# zonalStats function.\n",
    "topoCol = ee.ImageCollection([topo])\n",
    "\n",
    "# Define parameters for the zonalStats function.\n",
    "params = {\n",
    "    'bands': [0, 1],\n",
    "    'bandsRename': ['elevation', 'slope']\n",
    "}\n",
    "\n",
    "# Extract zonal statistics per point per image.\n",
    "ptsTopoStats = zonalStats(topoCol, ptsTopo, params)\n",
    "print('Topo zonal stats table', ptsTopoStats)\n",
    "\n",
    "# Display the layers on the map.\n",
    "Map.setCenter(-118.5957, 37.0775, 13)\n",
    "Map.addLayer(topoCol.select(0), {\n",
    "    'min': 2400,\n",
    "    'max': 4200\n",
    "}, 'Elevation')\n",
    "Map.addLayer(topoCol.select(1), {\n",
    "    'min': 0,\n",
    "    'max': 60\n",
    "}, 'Slope')\n",
    "Map.addLayer(pts, {\n",
    "    'color': 'purple'\n",
    "}, 'Points')\n",
    "Map.addLayer(ptsTopo, {\n",
    "    'color': 'yellow'\n",
    "}, 'Points w/ buffer')\n",
    "\n",
    "\n",
    "########################################\n",
    "\n",
    "# Example 2: MODIS\n",
    "\n",
    "ptsModis = pts.map(bufferPoints(50, True))\n",
    "\n",
    "# Load MODIS time series\n",
    "modisCol = ee.ImageCollection('MODIS/006/MOD09A1') \\\n",
    "    .filterDate('2015-01-01', '2020-01-01') \\\n",
    "    .filter(ee.Filter.calendarRange(183, 245, 'DAY_OF_YEAR'))\n",
    "\n",
    "# Define parameters for the zonalStats function.\n",
    "params = {\n",
    "    'reducer': ee.Reducer.median(),\n",
    "    'scale': 500,\n",
    "    'crs': 'EPSG:5070',\n",
    "    'bands': ['sur_refl_b01', 'sur_refl_b02', 'sur_refl_b06'],\n",
    "    'bandsRename': ['modis_red', 'modis_nir', 'modis_swir'],\n",
    "    'datetimeName': 'date',\n",
    "    'datetimeFormat': 'YYYY-MM-dd'\n",
    "}\n",
    "\n",
    "# Extract zonal statistics per point per image.\n",
    "ptsModisStats = zonalStats(modisCol, ptsModis, params)\n",
    "print('Limited MODIS zonal stats table', ptsModisStats.limit(50))\n",
    "\n",
    "########################################\n",
    "\n",
    "# Example 3: Landsat timeseries\n",
    "\n",
    "# Mask clouds from images and apply scaling factors.\n",
    "def maskScale(img):\n",
    "    qaMask = img.select('QA_PIXEL').bitwiseAnd(parseInt('11111',\n",
    "        2)).eq(0)\n",
    "    saturationMask = img.select('QA_RADSAT').eq(0)\n",
    "\n",
    "    # Apply the scaling factors to the appropriate bands.\n",
    "    def getFactorImg(factorNames):\n",
    "        factorList = img.toDictionary().select(factorNames) \\\n",
    "            .values()\n",
    "        return ee.Image.constant(factorList)\n",
    "    \n",
    "    scaleImg = getFactorImg(['REFLECTANCE_MULT_BAND_.'])\n",
    "    offsetImg = getFactorImg(['REFLECTANCE_ADD_BAND_.'])\n",
    "    scaled = img.select('SR_B.').multiply(scaleImg).add(\n",
    "    offsetImg)\n",
    "\n",
    "    # Replace the original bands with the scaled ones and apply the masks.\n",
    "    return img.addBands(scaled, None, True) \\\n",
    "        .updateMask(qaMask) \\\n",
    "        .updateMask(saturationMask)\n",
    "\n",
    "\n",
    "# Selects and renames bands of interest for Landsat OLI.\n",
    "def renameOli(img):\n",
    "    return img.select(\n",
    "        ['SR_B2', 'SR_B3', 'SR_B4', 'SR_B5', 'SR_B6', 'SR_B7'],\n",
    "        ['Blue', 'Green', 'Red', 'NIR', 'SWIR1', 'SWIR2'])\n",
    "\n",
    "\n",
    "# Selects and renames bands of interest for TM/ETM+.\n",
    "def renameEtm(img):\n",
    "    return img.select(\n",
    "        ['SR_B1', 'SR_B2', 'SR_B3', 'SR_B4', 'SR_B5', 'SR_B7'],\n",
    "        ['Blue', 'Green', 'Red', 'NIR', 'SWIR1', 'SWIR2'])\n",
    "\n",
    "\n",
    "# Prepares (cloud masks and renames) OLI images.\n",
    "def prepOli(img):\n",
    "    img = maskScale(img)\n",
    "    img = renameOli(img)\n",
    "    return img\n",
    "\n",
    "\n",
    "# Prepares (cloud masks and renames) TM/ETM+ images.\n",
    "def prepEtm(img):\n",
    "    img = maskScale(img)\n",
    "    img = renameEtm(img)\n",
    "    return img\n",
    "\n",
    "\n",
    "ptsLandsat = pts.map(bufferPoints(15, True))\n",
    "\n",
    "oliCol = ee.ImageCollection('LANDSAT/LC08/C02/T1_L2') \\\n",
    "    .filterBounds(ptsLandsat) \\\n",
    "    .map(prepOli)\n",
    "\n",
    "etmCol = ee.ImageCollection('LANDSAT/LE07/C02/T1_L2') \\\n",
    "    .filterBounds(ptsLandsat) \\\n",
    "    .map(prepEtm)\n",
    "\n",
    "tmCol = ee.ImageCollection('LANDSAT/LT05/C02/T1_L2') \\\n",
    "    .filterBounds(ptsLandsat) \\\n",
    "    .map(prepEtm)\n",
    "\n",
    "# Merge the sensor collections\n",
    "landsatCol = oliCol.merge(etmCol).merge(tmCol)\n",
    "\n",
    "# Define parameters for the zonalStats function.\n",
    "params = {\n",
    "    'reducer': ee.Reducer.max(),\n",
    "    'scale': 30,\n",
    "    'crs': 'EPSG:5070',\n",
    "    'bands': ['Blue', 'Green', 'Red', 'NIR', 'SWIR1', 'SWIR2'],\n",
    "    'bandsRename': ['ls_blue', 'ls_green', 'ls_red', 'ls_nir',\n",
    "        'ls_swir1', 'ls_swir2'\n",
    "    ],\n",
    "    'imgProps': ['SENSOR_ID', 'SPACECRAFT_ID'],\n",
    "    'imgPropsRename': ['img_id', 'satellite'],\n",
    "    'datetimeName': 'date',\n",
    "    'datetimeFormat': 'YYYY-MM-dd'\n",
    "}\n",
    "\n",
    "# Extract zonal statistics per point per image.\n",
    "ptsLandsatStats = zonalStats(landsatCol, ptsLandsat, params) \\\n",
    "    .filter(ee.Filter.NotNull(params.bandsRename))\n",
    "print('Limited Landsat zonal stats table', ptsLandsatStats.limit(50))\n",
    "\n",
    "Export.table.toAsset({\n",
    "    'collection': ptsLandsatStats,\n",
    "    'description': 'EEFA_export_Landsat_to_points',\n",
    "    'assetId': 'EEFA_export_values_to_points'\n",
    "})\n",
    "\n",
    "Export.table.toDrive({\n",
    "    'collection': ptsLandsatStats,\n",
    "    'folder': 'EEFA_outputs', # this will create a new folder if it doesn't exist\n",
    "    'description': 'EEFA_export_values_to_points',\n",
    "    'fileFormat': 'CSV'\n",
    "})\n",
    "\n",
    "#  -----------------------------------------------------------------------\n",
    "#  CHECKPOINT\n",
    "#  -----------------------------------------------------------------------\n",
    "\n",
    "\n",
    "\n"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Display the interactive map"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "Map"
   ]
  }
 ],
 "metadata": {
  "anaconda-cloud": {},
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.6.1"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 4
}
